/* MoreBalance includes */
#include "magic.h"

/* System includes */
#include <stdlib.h>
#include <stdio.h> /**/
#include <string.h>

/* Global variables */
struct magic_t
{
	int *ports;
	int numports;
	char ***nodes;
	int *numnodes;
	int *curnodes;
	int **traffic;
};

typedef struct magic_t magic;

static magic m;

/* Initialize all the magic */
void magic_initialize(conf *c)
{
	int i, j, k, pnumber, cnumber;

	m.ports = NULL;
	m.nodes = NULL;
	m.numnodes = NULL;
	m.curnodes = NULL;
	m.traffic = NULL;
	m.numports = 0;

	for(i = 0; i < c->clusternum; i++)
	{
		for(j = 0; j < c->cluster[i]->portsnum; j++)
		{
			/* Find port-based list in magic */
			pnumber = -1;
			for(k = 0; k < m.numports; k++)
			{
				if(m.ports[k] == c->cluster[i]->ports[j])
				{
					pnumber = k;
					break;
				}
			}

			if(pnumber < 0)
			{
				/* New port-based list to be created */
				pnumber = m.numports;
				m.ports = (int*)realloc(m.ports, (pnumber + 1) * sizeof(int));
				m.ports[pnumber] = c->cluster[i]->ports[j];

				m.curnodes = (int*)realloc(m.curnodes, (pnumber + 1) * sizeof(int));
				m.curnodes[pnumber] = 0;

				m.numnodes = (int*)realloc(m.numnodes, (pnumber + 1) * sizeof(int));
				m.numnodes[pnumber] = 0;

				m.nodes = (char***)realloc(m.nodes, (pnumber + 1) * sizeof(char**));
				m.nodes[pnumber] = NULL;

				m.traffic = (int**)realloc(m.traffic, (pnumber + 1) * sizeof(int*));
				m.traffic[pnumber] = NULL;

				m.numports++;
			}

			/* Fill port-based list with nodes */
			cnumber = m.numnodes[pnumber];
			for(k = 0; k < c->cluster[i]->nodesnum; k++)
			{
				m.nodes[pnumber] = (char**)realloc(m.nodes[pnumber], (cnumber + 1) * sizeof(char*));
				m.nodes[pnumber][cnumber] = strdup(c->cluster[i]->nodes[k]);

				m.traffic[pnumber] = (int*)realloc(m.traffic[pnumber], (cnumber + 1) * sizeof(int));
				m.traffic[pnumber][cnumber] = 0;

				cnumber++;
			}
			m.numnodes[pnumber] = cnumber;
		}
	}
}

/* Disable one node */
void magic_disable(const char *node)
{
}

/* Return a node depending on the ip and the port */
char *magic_node(conf *c, const char *ip, int port)
{
	int i, pnumber;

	if((!m.ports) || (!m.nodes)) return NULL;

	/* Find position for this port in list */
	i = 0;
	pnumber = -1;
	for(i = 0; i < m.numports; i++)
	{
		if(m.ports[i] == port)
		{
			pnumber = i;
			break;
		}
	}
	if(pnumber < 0) return NULL;

	/* Round Robin */
	if(m.nodes[pnumber][m.curnodes[pnumber]])
	{
		m.curnodes[pnumber]++;
		if(m.curnodes[pnumber] >= m.numnodes[pnumber]) m.curnodes[pnumber] = 0;
		return m.nodes[pnumber][m.curnodes[pnumber]];
	}

	return NULL;
}

/* Return a destination port depending on the ip and the port */
int magic_port(conf *c, const char *ip, int port)
{
	int i;
	for(i = 0; i < c->clusternum; i++)
	{
		if(c->cluster[i]->ports[0] == port)
		{
			if(!c->cluster[i]->destport) return port;
			return c->cluster[i]->destport;
		}
	}
	return port;
}

