/* MoreBalance includes */
#include "conf.h"
#include "server.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>

/* Child function: startup MoreBalance */
static void launch()
{
	conf c;

	c = conf_read("morebalance.conf");

	if(c.valid)
	{
		printf("Configuration loaded.\n");
		fflush(NULL);

		/*conf_dump(c);*/
		conf_optimize(&c);
		conf_dump(c);

		printf("Start the server.\n");
		fflush(NULL);

		server_start(c);
	}
	else
	{
		fprintf(stderr, "Error reading configuration file.\n");
		exit(-1);
	}
}

/* Main function: main safety loop */
int main(int argc, char *argv[])
{
	pid_t pid;
	int status;

	printf(">> Startup.\n");
	fflush(NULL);

	while(1)
	{
		pid = fork();
		switch(pid)
		{
			case -1:
				fprintf(stderr, "Couldn't fork.\n");
				exit(-1);
				break;
			case 0:
				launch();
				exit(-1);
				break;
			default:
				break;
		}

		printf(">> Loop.\n");
		fflush(NULL);

		wait(&status);
		if(!WIFSIGNALED(status)) break;

		fprintf(stderr, ">> Child crashed. Restarting.\n");
	}

	return 0;
}

