/* MoreBalance includes */
#include "server.h"
#include "magic.h"
#include "ipban.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <netinet/in.h>
#include <pthread.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

/* Globals */
conf global_c;

/* Connection structure */
struct conn_t
{
	int fd;
	char *hostname;
	int port;
	int destport;
};

/* Sub-subserver for each connection */
void *server_handle(void *arg)
{
	struct sockaddr_in sa;
	struct sockaddr_un su;
	struct hostent *h = NULL;
	int fd, fd2;
	int ret, ret2;
	ssize_t bytes;
	char buf[16384];
	long long rbytes, sbytes;
	struct conn_t *carg;

	ret = 1;

	carg = (struct conn_t*)arg;
	fd = carg->fd;

	if(carg->hostname[0] != '/')
	{
		h = gethostbyname(carg->hostname);
		if(!h)
		{
			fprintf(stderr, "Couldn't resolve %s.\n", carg->hostname);
			free(carg);
			pthread_exit(&ret);
		}

		fd2 = socket(AF_INET, SOCK_STREAM, 0);
	}
	else
	{
		fd2 = socket(AF_LOCAL, SOCK_STREAM, 0);
	}

	if(fd2 < 0)
	{
		fprintf(stderr, "Couldn't create socket.\n");
		free(carg);
		pthread_exit(&ret);
	}

	if(carg->hostname[0] != '/')
	{
		sa.sin_family = AF_INET;
		sa.sin_port = htons(carg->destport);
		memcpy(&sa.sin_addr, h->h_addr, h->h_length);

		ret2 = connect(fd2, (const struct sockaddr*)&sa, sizeof(sa));
	}
	else
	{
		su.sun_family = AF_LOCAL;
		strncpy(su.sun_path, carg->hostname, (sizeof(su.sun_path) - 1));
		/*unlink(carg->hostname);*/

		ret2 = connect(fd2, (const struct sockaddr*)&su, sizeof(su));
	}

	if(ret2 < 0)
	{
		fprintf(stderr, "Couldn't connect to %s:%i.\n", carg->hostname, carg->destport);
		close(fd);
		free(carg);
		pthread_exit(&ret);
	}

	fcntl(fd, F_SETFL, O_NONBLOCK);
	fcntl(fd2, F_SETFL, O_NONBLOCK);

	rbytes = 0;
	sbytes = 0;

	if(global_c.verbose) printf("--> Transmission starting: fd=%i, fd2=%i!\n", fd, fd2);
	fflush(NULL);
	while(1)
	{
		bytes = read(fd, buf, sizeof(buf));
		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				fprintf(stderr, "Read error on fd (client)!\n");
				break;
			}
		}
		else if(bytes > 0)
		{
			buf[bytes] = 0;
			if(global_c.verbose > 1) printf(":: -> %s", buf);
			fflush(NULL);
			bytes = write(fd2, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					fprintf(stderr, "Write error on fd2 (node)!\n");
					break;
				}
			}
			rbytes += bytes;
		}
		else if(bytes == 0)
		{
			if(global_c.verbose) fprintf(stderr, "broken pipe on fd (client)!\n");
			if(global_c.verbose > 1) printf("broken pipe on fd (client)!\n");
			fflush(NULL);
			break;
		}

		bytes = read(fd2, buf, sizeof(buf));
		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				fprintf(stderr, "Read error on fd2 (node)!\n");
				break;
			}
		}
		else if(bytes > 0)
		{
			buf[bytes] = 0;
			if(global_c.verbose > 1) printf(":: <- %s", buf);
			fflush(NULL);
			bytes = write(fd, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					fprintf(stderr, "Write error on fd (client)!\n");
					break;
				}
			}
			sbytes += bytes;
		}
		else if(bytes == 0)
		{
			if(global_c.verbose) fprintf(stderr, "broken pipe on fd2 (cluster node)!\n");
			if(global_c.verbose > 1) printf("broken pipe on fd2 (cluster node)!\n");
			fflush(NULL);
			break;
		}
	}

	close(fd);
	close(fd2);

	if(global_c.verbose)
	{
		printf("Sent bytes [cluster node -> client]: %qi\n", sbytes);
		printf("Read bytes [client -> cluster node]: %qi\n", rbytes);
	}
	fflush(NULL);

	free(carg);
	pthread_exit(&ret);
}

/* Threaded subserver */
void *server_sub(void *arg)
{
	int ret, ret2;
	int fd;
	int port, destport;
	struct sockaddr_in sa;
	struct sockaddr addr;
	int addrsize;
	unsigned short port2;
	char *host;
	int x;
	struct conn_t *carg;
	pthread_t id;
	pthread_attr_t attr;
	char ip[15];
	unsigned int ipx[5];

	ret = 1;

	port = *(int*)arg;
	if(global_c.verbose) printf("=> Port: %i\n", port);
	fflush(NULL);

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if(fd < 0)
	{
		fprintf(stderr, "Couldn't create file descriptor.\n");
		pthread_exit(&ret);
	}
	setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));

	sa.sin_family = AF_INET;
	sa.sin_addr.s_addr = htonl(INADDR_ANY);
	sa.sin_port = htons(port);
	ret = bind(fd, (const struct sockaddr*)&sa, sizeof(sa));
	if(ret)
	{
		fprintf(stderr, "Not all ports are available.\n");
		sleep(1);
		pthread_exit(&ret);
	}
	port2 = ntohs(sa.sin_port);

	ret = listen(fd, 16);
	if(ret)
	{
		fprintf(stderr, "Couldn't listen to socket.\n");
		pthread_exit(&ret);
	}

	while(1)
	{
		ret = accept(fd, &addr, &addrsize);
		if(ret < 0)
		{
			fprintf(stderr, "Couldn't accept connection.\n");
			pthread_exit(&ret);
		}

		if(global_c.verbose) printf("--> Connection on port %i!\n", port);
		fflush(NULL);

		ret2 = getpeername(ret, &addr, &addrsize);
		if(ret2)
		{
			fprintf(stderr, "Peer name not found.\n");
			pthread_exit(&ret);
		}

		sa = *(struct sockaddr_in*)&addr;
		ipx[0] = ntohl(sa.sin_addr.s_addr);
		ipx[1] = (ipx[0] >> 24) & 0xFF;
		ipx[2] = (ipx[0] >> 16) & 0xFF;
		ipx[3] = (ipx[0] >> 8) & 0xFF;
		ipx[4] = (ipx[0]) & 0xFF;
		snprintf(ip, sizeof(ip), "%u.%u.%u.%u", ipx[1], ipx[2], ipx[3], ipx[4]);
		ret2 = ipban_check(ip);
		if(ret2 == 1)
		{
			fprintf(stderr, "%s is blocked!\n", ip);
			close(fd);
			pthread_exit(&ret);
		}

		host = magic_node(&global_c, "", port);
		destport = magic_port(&global_c, host, port);

		carg = (struct conn_t*)malloc(sizeof(struct conn_t));
		carg->fd = ret;
		carg->hostname = host;
		carg->port = port;
		carg->destport = destport;

		/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&id, NULL, &server_handle, carg);
		pthread_attr_destroy(&attr);
	}

	pthread_exit(&ret);
}

/* Start all subservers */
void server_start(conf c)
{
	int i, j;
	pthread_t id;
	pthread_attr_t attr;

	if(!c.clusternum)
	{
		fprintf(stderr, "No clusters found.\n");
		exit(-1);
	}

	magic_initialize(&c);
	global_c = c;

	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
			/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
			pthread_attr_init(&attr);
			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
			pthread_create(&id, NULL, &server_sub, &(c.cluster[i]->ports[j]));
			pthread_attr_destroy(&attr);
		}
	}

	while(1)
		sleep(99999);
}

